#include "opencv2/opencv.hpp"
#include <vector>
#include <string>
#include <iostream>

void bw1 (const cv::Mat &src, cv::Mat &dst) 
{
	cv::cvtColor(src,dst,cv::COLOR_RGB2GRAY);
}

void bw2 (const cv::Mat &src, cv::Mat &dst) 
{
    assert(src.channels() == 3 && "color image required by bw3");
	std::vector<cv::Mat> channels;
	cv::split(src,channels);
	for (int i=0; i<3; i++) channels[i].convertTo(channels[i],CV_16UC1);
	dst = (channels[0]+channels[1]+channels[2])/3;
	dst.convertTo(dst,CV_8UC1);
}

void bw3 (const cv::Mat &src, cv::Mat &dst) 
{
	assert(src.type() == CV_8UC3 && "color image required by bw3");
	cv::Mat tmp(src.size(),CV_8UC1);
//	cv::Mat tmp(src.rows,src.cols,CV_8UC1);
//	cv::Mat tmp(Size(src.cols,src.rows),CV_8UC1);
	for (int y=0; y<src.rows; y++) {
		for (int x=0; x<src.cols; x++) {
			const cv::Vec3b &pixelsrc = src.at<cv::Vec3b>(y,x);
			uchar &pixeltmp = tmp.at<uchar>(y,x);
			pixeltmp = (pixelsrc[0]+pixelsrc[1]+pixelsrc[2])/3;
		}
	}
	dst = tmp;
}

void bw4 (cv::Mat &src, cv::Mat &dst) 
{
	assert(src.type() == CV_8UC3 && "color image required by bw4");
	cv::Mat tmp(src.size(),CV_8UC1);
	cv::MatIterator_<uchar> tmpit = tmp.begin<uchar>();
	for (cv::MatIterator_<cv::Vec3b> srcit = src.begin<cv::Vec3b>(); srcit != src.end<cv::Vec3b>(); srcit++, tmpit++) 
		*tmpit = ((*srcit)[0] + (*srcit)[1] + (*srcit)[2]) / 3;
	dst = tmp;
}

template<typename T>
void bw5 (cv::Mat &src, cv::Mat &dst) // slow
{
	assert(src.channels() == 3 && "three channels are required by bw5");
	cv::Mat tmp(src.size(),src.type()-2*8);
	for (int y=0; y<src.rows; y++) {
		for (int x=0; x<src.cols; x++) {
			cv::Vec<T,3> &pixelsrc = src.at<cv::Vec<T,3>>(y,x);
			T &pixeltmp = tmp.at<T>(y,x);
			pixeltmp = (pixelsrc[0]+pixelsrc[1]+pixelsrc[2])/3;
		}
	}
	dst = tmp;
}

void bw6 (const cv::Mat &src, cv::Mat &dst) // fast but dangerous (because of roi)
{
	assert(src.type() == CV_8UC3 && "color image required by bw2");
	const cv::Vec3b *srcp = src.ptr<cv::Vec3b>();
	cv::Mat tmp(src.size(),CV_8UC1);
	uchar *tmpp = tmp.ptr<uchar>();
	for (size_t i=0; i<src.total(); i++, srcp++, tmpp++) 
		*tmpp = ((*srcp)[0] + (*srcp)[1] + (*srcp)[2]) / 3;
	dst = tmp;
}

void bw7 (cv::Mat &src, cv::Mat &dst) // fast and safety
{
    assert(src.type() == CV_8UC3 && "color image required by bw2");
    cv::Mat tmp(src.size(),CV_8UC1);
    for (int y=0; y<src.rows; y++) {
        const cv::Vec3b *srcp = src.ptr<cv::Vec3b>(y); // == src.row(y).ptr<cv::Vec3b>()
        uchar *tmpp = tmp.ptr<uchar>(y); // == tmp.row(y).ptr<uchar>()
        for (int x=0; x<src.cols; x++) 
            tmpp[x] = (srcp[x][0] + srcp[x][1] + srcp[x][2]) / 3;
    }
    dst = tmp;
}

int main (int argc, char *argv[])
{
	cv::Mat image = cv::imread("../lena.jpg",cv::IMREAD_COLOR), gray;
	if (image.empty()) return -1;
    std::cout << "press 1, space, 2, space, ... 7, esc" << std::endl;
    gray = image;
	for (;;) {
		bool selected = true;
        cv::imshow("lena",gray);
        int key = cv::waitKey(0);
		switch (key) {
		case '1': 
			bw1(image,gray);
			break;
		case '2': 
			bw2(image,gray);
			break;
		case '3': 
			bw3(image,gray);
			break;
		case '4': 
			bw4(image,gray);
			break;
		case '5': 
			bw5<uchar>(image,gray);
			break;
        case '6': 
            bw6(image,gray);
            break;
        case '7': 
            bw7(image,gray);
            break;
        default: 
			selected = false;
		}
		if (!selected) gray = image;
        if (key == 27) break;
	}
	cv::destroyAllWindows();
	return 0;
}
